from typing import Annotated

from newsflash.svg.utils import Point
from newsflash.svg.box import Box, scale_point_to_box
from newsflash.svg.element import Element, TemplateParam


class Path(Element):
    template: tuple[str, str] = ("svg", "path.svg")
    points: Annotated[list[Point], TemplateParam()]
    path_length: Annotated[float, TemplateParam()] = 100.0


def build_path(
    points: list[Point],
    path_length: float = 100.0,
    box: Box | None = None,
    classes: list[str] = [],
    styles: list[str] = [],
    attributes: dict[str, str] = {},
) -> Path:
    if box is not None:
        points = [scale_point_to_box(p, box) for p in points]

    return Path(
        points=points,
        path_length=path_length,
        classes=classes,
        styles=styles,
        attributes=attributes,
    )



from typing import Annotated

from newsflash.svg.utils import Point
from newsflash.svg.box import Box, scale_point_to_box
from newsflash.svg.element import Element, TemplateParam


class Line(Element):
    template: tuple[str, str] = ("svg", "line.svg")
    from_pos: Annotated[Point, TemplateParam()]
    to_pos: Annotated[Point, TemplateParam()]
    stroke_color: Annotated[str, TemplateParam()] = "black"
    stroke_width: Annotated[float, TemplateParam()] = 1.0
    stroke_linecap: Annotated[str, TemplateParam()] = "round"
    path_length: Annotated[float, TemplateParam()] = 1.0


def build_line(
    from_pos: Point,
    to_pos: Point,
    stroke_color: str = "black",
    stroke_width: float = 1.0,
    stroke_linecap: str = "round",
    path_length: float = 1.0,
    box: Box | None = None,
    classes: list[str] = [],
    styles: list[str] = [],
    attributes: dict[str, str] = {},
) -> Line:
    if box is not None:
        from_pos = scale_point_to_box(from_pos, box)
        to_pos = scale_point_to_box(to_pos, box)

    return Line(
        from_pos=from_pos,
        to_pos=to_pos,
        stroke_color=stroke_color,
        stroke_width=stroke_width,
        stroke_linecap=stroke_linecap,
        path_length=path_length,
        classes=classes,
        styles=styles,
        attributes=attributes,
    )



from typing import Literal, Annotated

from fontTools.ttLib import TTFont

from newsflash.svg.element import Element, TemplateParam
from newsflash.svg.utils import Point, XCoor, YCoor
from newsflash.svg.box import Box, scale_point_to_box
from newsflash.svg.utils.fonts import get_text_width


type HorizontalAlign = Literal["left", "center", "right"]
type VerticalAlign = Literal["top", "middle", "default", "bottom"]


class Text(Element):
    template: tuple[str, str] = ("svg", "text.svg")
    pos: Annotated[Point, TemplateParam()]
    text: Annotated[str, TemplateParam()]
    width: Annotated[float, TemplateParam()]
    font_size: Annotated[int, TemplateParam()]

    @property
    def top(self) -> YCoor:
        return self.pos.y - self.font_size

    @property
    def right(self) -> XCoor:
        return self.pos.x + self.width

    @property
    def bottom(self) -> YCoor:
        return self.pos.y + self.font_size / 3

    @property
    def left(self) -> XCoor:
        return self.pos.x

    @property
    def height(self) -> float:
        return self.font_size * 4 / 3


def build_text(
    pos: Point,
    text: str,
    font: TTFont,
    font_size: int,
    horizontal_align: HorizontalAlign = "left",
    vertical_align: VerticalAlign = "default",
    box: Box | None = None,
    classes: list[str] = [],
    styles: list[str] = [],
    attributes: dict[str, str] = {},
) -> Text:
    if box is not None:
        pos = scale_point_to_box(pos, box)

    text_width = get_text_width(font, text, font_size)
    if horizontal_align == "center":
        pos.x = pos.x - text_width / 2
    elif horizontal_align == "right":
        pos.x = pos.x - text_width

    if vertical_align == "top":
        pos.y = pos.y + font_size
    elif vertical_align == "middle":
        pos.y = pos.y + font_size / 4
    elif vertical_align == "bottom":
        pos.y = pos.y - font_size / 4

    styles = styles + [f"font-size: {font_size}px;"]

    return Text(
        pos=pos,
        text=text,
        width=text_width,
        font_size=font_size,
        classes=classes,
        styles=styles,
        attributes=attributes,
    )



from typing import Type, TypeVar, Mapping
from pathlib import Path

from fastapi import FastAPI
from fastapi.staticfiles import StaticFiles
from fastapi.responses import HTMLResponse

from newsflash.widgets.widgets import Widget
from newsflash.templates.templates import template_registry
from newsflash.endpoints.callback import get_page_callback, get_callback_endpoint


W = TypeVar("W", bound=Widget)


class Page(Widget):
    path: str
    title: str


THEME_TO_STYLESHEET: dict[str, str] = {
    "newsflash": "/_newsflash/static/styles.css",
    "colorful": "/_newsflash/static/styles-2.css",
}


class App(FastAPI):
    pages: dict[str, Page]
    theme: str

    def __init__(
        self,
        pages: list[Page],
        template_folders: list[tuple[str, Path]],
        static_folders: list[tuple[str, Path]] = [],
        theme: str = "newsflash",
    ) -> None:
        super().__init__(docs_url=None, redoc_url=None, openapi_url=None)

        self.pages = {page.path: page for page in pages}

        for folder in template_folders:
            template_registry.register_template_folder(folder[0], folder[1])

        static_folders.append(
            ("/_newsflash/static", Path(__file__).parent / "assets" / "staticfiles")
        )
        self.theme = theme

        self._mount_static_folders(static_folders)
        self._build_page_endpoints()
        self._build_callback_endpoints()

    def get_widget(
        self,
        path: str,
        type: Type[W],
        id: str | None = None,
        body_params: Mapping[str, str] | None = None,
    ) -> W:
        page = self.pages[path].model_copy(copy=True, body_params=body_params)
        # assert id.startswith(f"{page.id}/"), (
        #     f"Widget id '{id}' not found on page '{page.id}'"
        # )
        if id is not None:
            id = id.removeprefix(f"{page.id}/")

        return page.get_child_widget(type=type, id=id)

    def _mount_static_folders(self, static_folders: list[tuple[str, Path]]) -> None:
        for mount_path, directory in static_folders:
            self.mount(
                mount_path,
                StaticFiles(directory=directory),
            )

    def _build_page_endpoints(self):
        for page_path, page in self.pages.items():
            assert page.template is not None, "Page template is not set."

            page_endpoint = get_page_callback(
                page=page,
                theme_stylesheet=THEME_TO_STYLESHEET.get(
                    self.theme, THEME_TO_STYLESHEET["newsflash"]
                ),
            )
            self.add_api_route(
                page_path, page_endpoint, methods=["GET"], response_class=HTMLResponse
            )

    def _build_callback_endpoints_for_widget(self, widget: Widget):
        full_path = widget.full_path

        if widget._callback_fn_name is not None:
            print(
                f"Setting up callback endpoint for widget {widget.id} ({widget._callback_fn_name}) at path /{full_path}"
            )
            self.add_api_route(
                f"/{full_path}",
                get_callback_endpoint(widget_id=widget.id, app=self),
                methods=["POST"],
                response_class=HTMLResponse,
            )

        if len(widget.children) > 0:
            for child in widget.children:
                self._build_callback_endpoints_for_widget(child)

    def _build_callback_endpoints(self):
        for page in self.pages.values():
            page.model_copy()
            for child in page.children:
                child.model_copy()
                self._build_callback_endpoints_for_widget(child)



from typing import Type

from .snapshot_test import SnapshotTest, AllSnapshotsPass
from .bar_chart import ascending, descending
from .line_chart import math as math_lines
from .histograms import distributions


def get_all_tests() -> dict[str, Type[SnapshotTest]]:
    all_tests = [
        ascending.AscendingBarChart5,
        ascending.AscendingBarChart20,
        descending.DescendingBarChart5,
        descending.DescendingBarChartLargeNumbers,
        math_lines.SineLine,
        math_lines.CosineLine,
        math_lines.SineLineWithCustomLabels,
        math_lines.SineLineWithCustomXRange,
        distributions.GaussianDistributionHistogram,
        distributions.BetaDistributionHistogram,
        distributions.UniformDistributionHistogram,
    ]

    return {test.name: test for test in all_tests}


def get_test_by_name(name: str) -> Type[SnapshotTest]:
    if name == "":
        return AllSnapshotsPass

    all_tests = get_all_tests()
    return all_tests[name]


def get_next_failing_test() -> Type[SnapshotTest]:
    for test in get_all_tests().values():
        if not test.passes_test():
            return test

    return AllSnapshotsPass



from typing import Any, Annotated

from newsflash.svg.element import Element, TemplateParam
from newsflash.svg.utils import Point, XCoor, YCoor
from newsflash.svg.box import (
    Box,
    scale_width_to_box,
    scale_height_to_box,
    scale_point_to_box,
)


class Rectangle(Element):
    template: tuple[str, str] = ("svg", "rect.svg")
    top_left: Annotated[Point, TemplateParam()]
    width: Annotated[float, TemplateParam()]
    height: Annotated[float, TemplateParam()]
    rounded: Annotated[float, TemplateParam()] = 0.0

    @property
    def top(self) -> YCoor:
        return self.top_left.y

    @property
    def right(self) -> XCoor:
        return self.top_left.x + self.width

    @property
    def bottom(self) -> YCoor:
        return self.top_left.y + self.height

    @property
    def left(self) -> XCoor:
        return self.top_left.x

    def get_additional_context(self) -> dict[str, Any]:
        return {
            "rx_ry": self.rounded * self.width,
        }


def build_rectangle(
    top_left: Point,
    width: float,
    height: float,
    rounded: float = 0.0,
    box: Box | None = None,
    classes: list[str] = [],
    styles: list[str] = [],
    attributes: dict[str, str] = {},
) -> Rectangle:
    if box is not None:
        top_left = scale_point_to_box(top_left, box)
        width = scale_width_to_box(width, box)
        height = scale_height_to_box(height, box)

    return Rectangle(
        top_left=top_left,
        width=width,
        height=height,
        rounded=rounded,
        classes=classes,
        styles=styles,
        attributes=attributes,
    )


def build_rectangle_from_bottom_center(
    bottom_center: Point,
    width: float,
    height: float,
    rounded: float = 0.0,
    box: Box | None = None,
    classes: list[str] = [],
    styles: list[str] = [],
    attributes: dict[str, str] = {},
) -> Rectangle:
    if box is not None:
        bottom_center = scale_point_to_box(bottom_center, box)
        width = scale_width_to_box(width, box)
        height = scale_height_to_box(height, box)

    top_left = Point(
        x=bottom_center.x - width / 2,
        y=bottom_center.y - height,
    )

    return Rectangle(
        top_left=top_left,
        width=width,
        height=height,
        rounded=rounded,
        classes=classes,
        styles=styles,
        attributes=attributes,
    )


def build_rectangle_from_top_center(
    top_center: Point,
    width: float,
    height: float,
    rounded: float = 0.0,
    box: Box | None = None,
    classes: list[str] = [],
    styles: list[str] = [],
    attributes: dict[str, str] = {},
) -> Rectangle:
    if box is not None:
        top_center = scale_point_to_box(top_center, box)
        width = scale_width_to_box(width, box)
        height = scale_height_to_box(height, box)

    top_left = Point(
        x=top_center.x - width / 2,
        y=top_center.y,
    )

    return Rectangle(
        top_left=top_left,
        width=width,
        height=height,
        rounded=rounded,
        classes=classes,
        styles=styles,
        attributes=attributes,
    )


def build_background_rectangle(
    box: Box,
    classes: list[str] = [],
    styles: list[str] = [],
    attributes: dict[str, str] = {},
) -> Rectangle:
    box = box.apply_padding()
    return build_rectangle(
        top_left=Point(x=box.left_x_value, y=box.top_y_value),
        width=abs(box.right_x_value - box.left_x_value),
        height=abs(box.bottom_y_value - box.top_y_value),
        box=box,
        classes=classes,
        styles=styles,
        attributes=attributes,
    )



from typing import Any, Annotated

from pytest import raises
from pydantic import ValidationError

from newsflash.widgets.widgets import Widget, BodyParam


class DummyWidget(Widget):
    id: str = "dummy-widget"
    test_value_a: Annotated[str, BodyParam()] = ""
    test_value_b: Annotated[float, BodyParam()] = 0.0

    _values_from_request: list[str] = ["test_value_a", "test_value_b"]
    _callback_fn_name: str = "dummy_callback"

    def dummy_callback(self) -> list[Widget]: ...


def test_widget_set_value_from_request():
    body_params: dict[str, Any] = {
        "dummy-widget-test_value_a": "Hello, World!",
        "dummy-widget-test_value_b": 3.14,
    }
    dummy_widget = DummyWidget().model_copy(
        body_params=body_params,
    )

    assert dummy_widget.test_value_a == "Hello, World!"
    assert dummy_widget.test_value_b == 3.14


def test_widget_set_value_no_attribute():
    body_params: dict[str, Any] = {
        "dummy-widget-test_value_a": "Hello, World!",
        "dummy-widget-test_value_c": 3.14,
    }
    dummy_widget = DummyWidget().model_copy(
        body_params=body_params,
    )

    assert dummy_widget.test_value_a == "Hello, World!"
    assert dummy_widget.test_value_b == 0.0
    assert not hasattr(dummy_widget, "test_value_c")


# TODO: design expected behavior for missing keys
# def test_widget_set_value_from_request_missing_key():
#     dummy_widget = DummyWidget()
#     request_data: dict[str, Any] = {
#         "dummy-widget-test_value_a": "Hello, World!",
#         # "dummy-widget-test_value_b" is missing
#     }

#     with raises(AssertionError) as e:
#         dummy_widget._set_value_from_request("test_value_b", request_data)

#     assert str(e.value) == "No value provided for key 'test_value_b'"


def test_widget_set_value_wrong_type():
    body_params: dict[str, Any] = {
        "dummy-widget-test_value_a": 100,
        "dummy-widget-test_value_b": 3.14,
    }

    with raises(ValidationError) as e:
        DummyWidget().model_copy(
            body_params=body_params,
        )


def test_get_widget_callback_fn():
    dummy_widget = DummyWidget()
    callback_fn = dummy_widget._get_callback_fn()
    assert callback_fn == dummy_widget.dummy_callback


def test_get_widget_callback_none():
    class NoCallbackWidget(Widget):
        id: str = "no-callback-widget"
        _callback_fn_name: str | None = None

    no_callback_widget = NoCallbackWidget()

    callback_fn = no_callback_widget._get_callback_fn()
    assert callback_fn is None


def test_get_widget_callback_fn_no_callback():
    class NoCallbackWidget(Widget):
        id: str = "no-callback-widget"
        _callback_fn_name: str = "non_existent_callback"

    with raises(AssertionError) as e:
        NoCallbackWidget().model_copy()

    assert "Widget has no callback function 'non_existent_callback'" in str(e.value)



from newsflash.svg.elements.path import Path, build_path
from newsflash.svg.utils import Point
from newsflash.svg.box import Box


def test_build_path():
    points = [Point(x=0, y=0), Point(x=50, y=100), Point(x=75, y=50), Point(x=100, y=0)]

    path = build_path(points=points)
    expected = Path(
        points=points,
        path_length=100.0,
        classes=[],
        styles=[],
        attributes={},
    )

    assert path == expected


def test_build_path_with_box():
    points = [
        Point(x=0, y=0),
        Point(x=0.5, y=1.0),
        Point(x=0.7, y=0.5),
        Point(x=1.0, y=0.0),
    ]
    box = Box(top=0, right=200, bottom=200, left=0)

    path = build_path(points=points, box=box)
    expected_points = [
        Point(x=0, y=0),
        Point(x=100, y=200),
        Point(x=140, y=100),
        Point(x=200, y=0),
    ]
    expected = Path(
        points=expected_points,
        path_length=100.0,
        classes=[],
        styles=[],
        attributes={},
    )

    assert path == expected



from pathlib import Path
import random

from newsflash.widgets import Histogram
from ..snapshot_test import SnapshotTest


class GaussianDistributionHistogram(SnapshotTest):
    name: str = "Gaussian Distribution Histogram"
    path_to_rendered: Path = Path(__file__).parent / "rendered" / "gauss.html"
    description: str = """
        Expected to see a histogram representing a Gaussian distribution with 20 bins.
        The data consists of 1000 samples drawn from a normal distribution with mean 10 and standard deviation 2.
        The width of the chart is 600 pixels.
    """

    @staticmethod
    def render() -> str:
        chart = Histogram(
            width=600,
            height=400,
        )

        chart.title = "Gaussian Distribution Histogram"
        random.seed(42)

        chart.set_values(
            values=[random.gauss(10, 2) for _ in range(9000)],
            num_bins=20,
        )

        return chart._render_update()


class BetaDistributionHistogram(SnapshotTest):
    name: str = "Beta Distribution Histogram"
    path_to_rendered: Path = Path(__file__).parent / "rendered" / "beta.html"
    description: str = """
        Expected to see a histogram representing a Beta distribution with 20 bins.
        The data consists of 1000 samples drawn from a Beta distribution with parameters alpha=2 and beta=5.
        The width of the chart is 600 pixels.
    """

    @staticmethod
    def render() -> str:
        chart = Histogram(
            width=600,
            height=400,
        )

        chart.title = "Beta Distribution Histogram"
        random.seed(42)

        chart.set_values(
            values=[random.betavariate(2, 5) for _ in range(9000)],
            num_bins=20,
        )

        return chart._render_update()


class UniformDistributionHistogram(SnapshotTest):
    name: str = "Uniform Distribution Histogram"
    path_to_rendered: Path = Path(__file__).parent / "rendered" / "uniform.html"
    description: str = """
        Expected to see a histogram representing a Uniform distribution with 20 bins.
        The data consists of 1000 samples drawn from a Uniform distribution between 0 and 1.
        The width of the chart is 600 pixels.
    """

    @staticmethod
    def render() -> str:
        chart = Histogram(
            width=600,
            height=400,
        )

        chart.title = "Uniform Distribution Histogram"
        random.seed(42)

        chart.set_values(
            values=[random.uniform(0, 1) for _ in range(9000)],
            num_bins=20,
        )

        return chart._render_update()



from .line import Line, build_line
from .rectangle import (
    Rectangle,
    build_rectangle,
    build_rectangle_from_bottom_center,
    build_rectangle_from_top_center,
    build_background_rectangle,
)
from .text import Text, build_text
from .path import Path, build_path
from .svg import SVG


__all__ = [
    "Line",
    "build_line",
    "Rectangle",
    "build_rectangle",
    "build_rectangle_from_bottom_center",
    "build_rectangle_from_top_center",
    "build_background_rectangle",
    "Text",
    "build_text",
    "Path",
    "build_path",
    "SVG",
]



from pytest import raises

from newsflash.widgets import Select, Input, TextArea, Button


def test_select_init_with_selected():
    select_widget = Select(
        options=["Option 1", "Option 2", "Option 3"], selected="Option 2"
    )

    assert select_widget.selected == "Option 2"


def test_select_init_default():
    # Test with default function
    def default_option():
        return "Option 2"

    select_widget = Select(
        options=["Option 1", "Option 2", "Option 3"], default=default_option
    ).model_copy()

    assert select_widget.selected == "Option 2"


def test_select_init_no_default():
    # Test without default function
    select_widget = Select(options=["Option 1", "Option 2", "Option 3"]).model_copy()

    assert (
        select_widget.selected == "Option 1"
    )  # First option should be selected by default


def test_select_no_options():
    select = Select(options=[])
    with raises(ValueError, match="Select widget has no options to select from."):
        select.model_copy()


def test_select_default_on_select():
    select_widget = Select(options=["Option 1", "Option 2", "Option 3"])

    # Simulate selecting an option
    select_widget.selected = "Option 2"
    result = select_widget.on_select()

    assert isinstance(result, list)  # Ensure it returns a list
    assert len(result) == 0  # No widgets should be returned by default


def test_input_default_on_input():
    input_widget = Input(value="Test input")

    # Simulate input event
    result = input_widget.on_input()

    assert isinstance(result, list)  # Ensure it returns a list
    assert len(result) == 0  # No widgets should be returned by default


def test_textarea_default_on_input():
    textarea_widget = TextArea(value="Test textarea", rows=5)

    # Simulate input event
    result = textarea_widget.on_input()

    assert isinstance(result, list)  # Ensure it returns a list
    assert len(result) == 0  # No widgets should be returned by default


def test_button_default_on_click():
    button_widget = Button(label="Submit")

    # Simulate button click
    result = button_widget.on_click()

    assert isinstance(result, list)  # Ensure it returns a list
    assert len(result) == 0  # No widgets should be returned by default



from typing import Annotated

from .widgets import Widget
from newsflash.svg.element import TemplateParam


class HTML(Widget):
    template: tuple[str, str] = ("widgets", "html.html")
    html_content: Annotated[str, TemplateParam()] = ""
    


from typing import Any, Annotated

from pydantic import BaseModel, ConfigDict

from newsflash.templates.templates import template_registry


class TemplateParam:
    template_param_name: str | None = None

    def __init__(self, template_param_name: str | None = None) -> None:
        self.template_param_name = template_param_name

    def get_template_param_name(self) -> str | None:
        return self.template_param_name


class Element(BaseModel):
    template: tuple[str, str] | None = None
    id: Annotated[str, TemplateParam()] = ""

    classes: Annotated[list[str], TemplateParam()] = []
    styles: Annotated[list[str], TemplateParam()] = []
    attributes: Annotated[dict[str, str], TemplateParam()] = {}

    model_config = ConfigDict(arbitrary_types_allowed=True, validate_assignment=True)

    def _get_all_template_params(self) -> dict[str, Any]:
        template_params: dict[str, Any] = {}

        for k, v in self.__class__.model_fields.items():
            if len(v.metadata) == 0:
                continue

            if p := next((m for m in v.metadata if isinstance(m, TemplateParam)), None):
                template_param_name = p.get_template_param_name() or k
                template_params[template_param_name] = getattr(self, k)

        return template_params

    def get_additional_context(self) -> dict[str, Any]:
        return {}

    def render(self, additional_context: dict[str, Any] = {}) -> str:
        context = self._get_all_template_params()
        context.update(self.get_additional_context())
        context.update(additional_context)

        assert self.template is not None, (
            f"Element with ID '{self.id}' has no template defined"
        )

        template = template_registry.get_template(
            template_folder=self.template[0], template_name=self.template[1]
        )

        rendered = template.render(context)
        return rendered


class ElementGroup(Element):
    elements: list[Element] = []

    def append(self, element: Element) -> None:
        self.elements.append(element)

    def extend(self, other: "ElementGroup") -> None:
        self.elements.extend(other.elements)

    def get_additional_context(self) -> dict[str, Any]:
        return {
            "content": "\n    ".join(element.render() for element in self.elements),
        }



from pathlib import Path

from jinja2 import Template
from fastapi.templating import Jinja2Templates


ROOT_TEMPLATE_DIR = Path(__file__).parent


class TemplateRegistry:
    def __init__(self) -> None:
        self._registry: dict[str, Jinja2Templates] = {}

    def register_template_folder(self, folder_name: str, directory: Path) -> None:
        self._registry[folder_name] = Jinja2Templates(directory=directory)

    def get_template(self, template_folder: str, template_name: str) -> Template:
        templates = self._registry[template_folder]
        return templates.get_template(template_name)


template_registry = TemplateRegistry()
template_registry.register_template_folder("widgets", ROOT_TEMPLATE_DIR / "widgets")
template_registry.register_template_folder("svg", ROOT_TEMPLATE_DIR / "svg")



from pytest import fixture

from newsflash.svg.element import Element, ElementGroup


class DummyElement(Element):
    element_number: int


@fixture
def element1() -> Element:
    return DummyElement(element_number=1)


@fixture
def element2() -> Element:
    return DummyElement(element_number=2)


@fixture
def element3() -> Element:
    return DummyElement(element_number=3)


def test_append(element1, element2):
    group = ElementGroup()

    group.append(element1)
    assert group.elements == [element1]

    group.append(element2)
    assert group.elements == [element1, element2]


def test_extend(element1, element2, element3):
    group1 = ElementGroup()
    group1.append(element1)

    group2 = ElementGroup()
    group2.append(element2)
    group2.append(element3)

    group1.extend(group2)

    assert group1.elements == [element1, element2, element3]



from pathlib import Path

from pytest import fixture, raises

from newsflash import App, Page
from newsflash.widgets.widgets import Widget


class DummyWidgetA(Widget):
    id: str = "widget_a"


class DummyWidgetB(Widget):
    id: str = "widget_b"


class DummyWidgetC(Widget):
    id: str = "widget_c"


class DummyPage(Page):
    id: str = "test-page"
    path: str = "/test-page"
    title: str = "Test Page"
    template: tuple[str, str] = ("templates", "test_page.html")

    def compose(self) -> list[Widget]:
        return [DummyWidgetA(), DummyWidgetB()]


class DummyPageSameWidget(Page):
    id: str = "test-page"
    path: str = "/test-page"
    title: str = "Test Page"
    template: tuple[str, str] = ("templates", "test_page.html")

    def compose(self) -> list[Widget]:
        return [DummyWidgetA(), DummyWidgetA(id="widget_a_2")]


# @fixture
# def test_page() -> Page:
#     return Page(
#         id="test-page",
#         path="/test-page",
#         title="Test Page",
#         template=("templates", "test_page.html"),
#         children=[DummyWidgetA(), DummyWidgetB()],
#     )


@fixture
def test_app() -> App:
    test_page = DummyPage()
    app = App(
        pages=[test_page],
        template_folders=[("templates", Path.cwd() / "dummy" / "path")],
    )
    return app


@fixture
def test_app_same_widget() -> App:
    test_page = DummyPageSameWidget()
    app = App(
        pages=[test_page],
        template_folders=[("templates", Path.cwd() / "dummy" / "path")],
    )
    return app


def test_get_widget_by_type(
    test_app: App,
):
    widget_a = test_app.get_widget(
        path="/test-page",
        type=DummyWidgetA,
    )
    assert isinstance(widget_a, DummyWidgetA)
    assert widget_a.id == "widget_a"

    widget_b = test_app.get_widget(path="/test-page", type=DummyWidgetB)
    assert isinstance(widget_b, DummyWidgetB)
    assert widget_b.id == "widget_b"


def test_query_one_with_id(
    test_app: App,
):
    widget_a = test_app.get_widget(
        path="/test-page",
        type=Widget,
        id="widget_a",
    )
    assert isinstance(widget_a, DummyWidgetA)
    assert widget_a.id == "widget_a"


def test_query_one_multiple_found(
    test_app_same_widget: App,
):
    test_app_same_widget.pages["/test-page"].children.append(
        DummyWidgetA(id="widget_a_2")
    )

    with raises(ValueError) as e:
        test_app_same_widget.get_widget(
            path="/test-page",
            type=DummyWidgetA,
        )

    assert "specify an id" in str(e.value).lower()


def test_query_one_not_found(
    test_app: App,
):
    with raises(ValueError) as e:
        test_app.get_widget(
            path="/test-page",
            type=DummyWidgetC,
        )

    assert "no widgets" in str(e.value).lower()



from typing import Annotated

from pydantic import BaseModel

from newsflash.svg.element import TemplateParam
from .widgets import Widget


class ValueDisplay(Widget):
    template: tuple[str, str] = ("widgets", "value_display.html")
    label: Annotated[str, TemplateParam()] = ""
    value: Annotated[str, TemplateParam()] = ""


class Paragraph(Widget):
    template: tuple[str, str] = ("widgets", "paragraph.html")
    text: Annotated[str, TemplateParam()] = ""


class Notification(BaseModel):
    message: str = ""
    level: str
    duration: int  # duration in milliseconds


class Notifications(Widget):
    id: Annotated[str, TemplateParam()] = "newsflash-notifications-container"
    template: tuple[str, str] = ("widgets", "notifications.html")
    notifications: Annotated[list[Notification], TemplateParam()] = []  # List of notifications to display

    def push(self, message: str, level: str = "info", duration: int = 5000) -> None:
        notification = Notification(
            message=message,
            level=level,
            duration=duration,
        )
        self.notifications.append(notification)



from typing import Any, Callable, Annotated, Mapping, Self

from newsflash.svg.element import TemplateParam
from .widgets import Widget, BodyParam


class Input(Widget):
    template: tuple[str, str] = ("widgets", "input.html")
    value: Annotated[str | None, BodyParam(), TemplateParam()] = None
    default: Callable[[], str] | None = None
    autofocus: Annotated[bool, TemplateParam()] = False
    type: Annotated[str, TemplateParam()] = "text"
    placeholder: Annotated[str, TemplateParam()] = ""

    _callback_fn_name: str = "on_input"

    def model_copy(
        self,
        *,
        copy: bool = False,
        update: Mapping[str, Any] | None = None,
        query_params: Mapping[str, list[str]] | None = None,
        body_params: Mapping[str, Any] | None = None,
        parent: Widget | None = None,
    ) -> Self:
        new_instance = super().model_copy(
            copy=copy,
            update=update,
            query_params=query_params,
            body_params=body_params,
            parent=parent,
        )
        if new_instance.value is None:
            if new_instance.default is not None:
                new_instance.value = new_instance.default()
            else:
                new_instance.value = ""

        return new_instance

    def on_input(self, *args, **kwargs) -> list[Widget]:
        """Event handler for input events."""
        return []


class TextArea(Input):
    template: tuple[str, str] = ("widgets", "textarea.html")
    rows: Annotated[int, TemplateParam()] = 7
    spellcheck: Annotated[bool, TemplateParam()] = False


class Select(Widget):
    template: tuple[str, str] = ("widgets", "select.html")
    options: Annotated[list[str], TemplateParam()] = ["-"]
    selected: Annotated[str | None, BodyParam(), TemplateParam()] = None
    default: Callable[[], str] | None = None

    _callback_fn_name: str = "on_select"

    def model_copy(
        self,
        *,
        copy: bool = False,
        update: Mapping[str, Any] | None = None,
        query_params: Mapping[str, list[str]] | None = None,
        body_params: Mapping[str, Any] | None = None,
        parent: Widget | None = None,
    ) -> Self:
        new_instance = super().model_copy(
            copy=copy,
            update=update,
            query_params=query_params,
            body_params=body_params,
            parent=parent,
        )

        if new_instance.selected is None:
            if new_instance.default is not None:
                new_instance.selected = new_instance.default()
            elif len(new_instance.options) > 0:
                new_instance.selected = new_instance.options[0]
            else:
                raise ValueError("Select widget has no options to select from.")

        return new_instance

    def on_select(self, *args, **kwargs) -> list[Widget]:
        """Event handler for select events."""
        return []


class Button(Widget):
    template: tuple[str, str] = ("widgets", "button.html")
    label: Annotated[str, TemplateParam()] = "Click Me"
    disabled: Annotated[bool, TemplateParam()] = False
    classes: Annotated[list[str], TemplateParam()] = ["newsflash-button"]

    _callback_fn_name: str = "on_click"

    def on_click(self, *args, **kwargs) -> list[Widget]:
        """Event handler for button click events."""
        return []



from pathlib import Path

from newsflash.widgets import BarChart
from ..snapshot_test import SnapshotTest


class DescendingBarChart5(SnapshotTest):
    name: str = "Descending Bar Chart 5"
    path_to_rendered: Path = Path(__file__).parent / "rendered" / "descending-5.html"
    description: str = """
        Expected to see five bars labelled A to E with heights decreasing from 
        left to right in steps of 30. The width of the chart is 600 pixels.
    """

    @staticmethod
    def render() -> str:
        chart = BarChart(
            width=600,
            height=400,
        )

        chart.title = "Ascending Bar Chart"

        chart.set_values(
            values=[150, 120, 90, 60, 30],
            labels=["A", "B", "C", "D", "E"],
            step=30,
        )

        return chart._render_update()


class DescendingBarChartLargeNumbers(SnapshotTest):
    name: str = "Descending Bar Chart Large Numbers"
    path_to_rendered: Path = (
        Path(__file__).parent / "rendered" / "descending-large-numbers.html"
    )
    description: str = """
        Expected to see five bars labelled A to E with heights decreasing from 
        left to right in steps of 1,000,000. The left starting point is 10,000,000.
        The width of the chart is 600 pixels.
        The y-axis is expected to scale appropriately to accommodate the large numbers,
        and display a multiplier of x1,000.000 above the y-axis.
    """

    @staticmethod
    def render() -> str:
        chart = BarChart(
            width=600,
            height=400,
        )

        chart.title = "Large Numbers"

        chart.set_values(
            values=[10_000_000, 9_000_000, 8_000_000, 7_000_000, 6_000_000],
            labels=["A", "B", "C", "D", "E"],
            step=2_000_000,
        )

        return chart._render_update()



from newsflash.svg.utils import Point
from newsflash.svg.box import Box
from newsflash.svg.elements import build_line


def test_build_line():
    line = build_line(
        from_pos=Point(x=10, y=20),
        to_pos=Point(x=200, y=100),
        stroke_color="red",
        stroke_width=2.0,
        stroke_linecap="square",
        path_length=5.0,
    )

    assert line.from_pos.x == 10
    assert line.from_pos.y == 20
    assert line.to_pos.x == 200
    assert line.to_pos.y == 100
    assert line.stroke_color == "red"
    assert line.stroke_width == 2.0
    assert line.stroke_linecap == "square"
    assert line.path_length == 5.0


def test_build_line_with_box():
    box = Box(
        top=0,
        right=300,
        bottom=300,
        left=0,
    )

    line = build_line(
        from_pos=Point(x=0.1, y=0.2),
        to_pos=Point(x=0.8, y=0.5),
        stroke_color="blue",
        stroke_width=3.0,
        box=box,
    )

    assert line.from_pos.x == 30  # 0.1 scaled to box width 300
    assert line.from_pos.y == 60  # 0.2 scaled to box height 300
    assert line.to_pos.x == 240  # 0.8 scaled to box width 300
    assert line.to_pos.y == 150  # 0.5 scaled to box height 300
    assert line.stroke_color == "blue"
    assert line.stroke_width == 3.0
    assert line.stroke_linecap == "round"  # default value
    assert line.path_length == 1.0  # default value



from newsflash.svg.element import Element
from newsflash.svg.elements import SVG


class DummyElement(Element):
    def render(self) -> str:
        return "<dummy></dummy>"


def test_svg_initialization():
    svg = SVG(width=200, height=100)
    assert svg.width == 200
    assert svg.height == 100
    assert svg.elements == []


def test_svg_append():
    svg = SVG(width=200, height=100)
    element = DummyElement()
    svg.append(element)
    assert len(svg.elements) == 1
    assert element in svg.elements


def test_svg_render():
    svg = SVG(id="test-svg", width=200, height=100)
    element = DummyElement()
    svg.append(element)

    result = svg.render()
    expected = """<svg
    xmlns="http://www.w3.org/2000/svg" 
    id="test-svg"
    class="" 
    width="200.0"
    height="100.0"
    viewBox="0 0 200.0 100.0"
    style=""
>
    <dummy></dummy>
</svg>"""
    assert result == expected


def test_svg_render_swap_oob():
    svg = SVG(id="test-svg", width=200, height=100)
    svg.hx_swap_oob = True
    element = DummyElement()
    svg.append(element)

    result = svg.render()
    expected = """<svg
    xmlns="http://www.w3.org/2000/svg" 
    id="test-svg"
    class="" 
    width="200.0"
    height="100.0"
    viewBox="0 0 200.0 100.0"
    style=""
    hx-swap-oob="true"
>
    <dummy></dummy>
</svg>"""
    assert result == expected



from fontTools.ttLib import TTFont

from newsflash.svg.box import Box
from newsflash.svg.utils.fonts import lora, get_text_width
from newsflash.svg.element import ElementGroup
from newsflash.svg.utils.fonts import font_size_to_height

from .axes import AxesConfig

from .components import (
    build_title_box,
    build_title_text,
    build_multipler_text,
    build_y_axis_box,
    build_y_axis,
    build_x_axis_box,
    build_x_axis,
    build_chart_box,
    build_horizontal_grid_lines,
)


def build_xy_chart(
    axes: AxesConfig,
    width: float,
    height: float,
    title: str,
    x_padding: float = 0.0,
    font: TTFont = lora,
    title_font_size: int = 32,
    label_font_size: int = 16,
) -> tuple[ElementGroup, Box]:
    xy_chart_elements = ElementGroup()

    # Title
    if title == "":
        title_font_size = label_font_size
    title_box = build_title_box(
        width=width, title_font_size=title_font_size, label_font_size=label_font_size
    )
    title_text = build_title_text(box=title_box, title=title, font_size=title_font_size)
    xy_chart_elements.append(title_text)

    if axes.y.multiplier != 1:
        multiplier_text = build_multipler_text(
            multiplier=axes.y.multiplier,
            font_size=label_font_size,
            box=title_box,
        )
        xy_chart_elements.append(multiplier_text)

    title_spacing_box = Box(
        top=title_box.bottom,
        right=title_box.right,
        bottom=title_box.bottom + font_size_to_height(label_font_size),
        left=title_box.left,
    )

    # Y-Axis
    max_label_width = max(
        [
            get_text_width(font=font, text=str(label), font_size=label_font_size)
            for label in axes.y.labels_as_str
        ]
    )
    y_axis_box = build_y_axis_box(
        min_value=axes.y.min_value,
        max_value=axes.y.max_value,
        title_box=title_spacing_box,
        max_label_width=max_label_width,
        svg_height=height,
        label_font_size=label_font_size,
    )
    y_axis = build_y_axis(
        y_labels=[str(label) for label in axes.y.labels_as_str],
        y_label_positions=axes.y.label_positions,
        font_size=label_font_size,
        y_axis_box=y_axis_box,
    )
    xy_chart_elements.extend(y_axis)

    # X-Axis
    x_axis_box = build_x_axis_box(
        svg_height=height,
        svg_width=width,
        font_size=label_font_size,
        y_axis_box=y_axis_box,
        min_x_axis_value=axes.x.min_value,
        max_x_axis_value=axes.x.max_value,
        padding_left=x_padding,
        padding_right=x_padding,
    )
    x_axis = build_x_axis(
        labels=axes.x.labels_as_str,
        label_positions=axes.x.label_positions,
        font_size=label_font_size,
        x_axis_box=x_axis_box,
    )
    xy_chart_elements.extend(x_axis)

    # Chart
    chart_box = build_chart_box(
        svg_width=width,
        min_x_axis_value=axes.x.min_value,
        max_x_axis_value=axes.x.max_value,
        min_y_axis_value=axes.y.min_value,
        max_y_axis_value=axes.y.max_value,
        title_box=title_spacing_box,
        x_axis_box=x_axis_box,
        y_axis_box=y_axis_box,
        padding_left=x_padding,
        padding_right=x_padding,
    )
    horizontal_grid_lines = build_horizontal_grid_lines(
        y_labels=axes.y.label_positions,
        chart_box=chart_box,
        padding_left=x_padding,
        padding_right=x_padding,
    )
    xy_chart_elements.extend(horizontal_grid_lines)

    return xy_chart_elements, chart_box



from pytest import fixture

from newsflash.svg.elements import (
    Rectangle,
    build_rectangle,
    build_rectangle_from_bottom_center,
    build_background_rectangle,
)
from newsflash.svg.utils import Point
from newsflash.svg.box import Box


def test_build_rectangle():
    rect = build_rectangle(
        top_left=Point(x=10, y=20),
        width=200,
        height=100,
        rounded=5,
        classes=["test-rect"],
    )

    assert rect.top_left.x == 10
    assert rect.top_left.y == 20
    assert rect.width == 200
    assert rect.height == 100
    assert rect.rounded == 5
    assert "test-rect" in rect.classes


def test_build_rectangle_with_box():
    box = Box(
        top=0,
        right=200,
        bottom=200,
        left=0,
    )

    rect = build_rectangle(
        top_left=Point(x=0.5, y=0.5),
        width=0.5,
        height=0.5,
        rounded=5,
        classes=["boxed-rect"],
        box=box,
    )

    assert rect.top_left.x == 100  # 0.5 scaled to box width 200
    assert rect.top_left.y == 100  # 0.5 scaled to box height 200
    assert rect.width == 100  # 0.5 scaled to box width 200
    assert rect.height == 100  # 0.5 scaled to box height 200
    assert rect.rounded == 5
    assert "boxed-rect" in rect.classes


def test_build_rectangle_from_bottom_center():
    rect = build_rectangle_from_bottom_center(
        bottom_center=Point(x=150, y=300),
        width=200,
        height=100,
        rounded=10,
        classes=["bottom-center-rect"],
    )

    assert rect.top_left.x == 50  # 150 - 200/2
    assert rect.top_left.y == 200  # 300 - 100
    assert rect.width == 200
    assert rect.height == 100
    assert rect.rounded == 10
    assert "bottom-center-rect" in rect.classes


def test_build_rectangle_from_bottom_center_with_box():
    box = Box(
        top=0,
        right=400,
        bottom=400,
        left=0,
    )

    rect = build_rectangle_from_bottom_center(
        bottom_center=Point(x=0.75, y=0.75),
        width=0.5,
        height=0.25,
        rounded=10,
        classes=["boxed-bottom-center-rect"],
        box=box,
    )

    assert rect.top_left.x == 200  # 300 - 200/2
    assert rect.top_left.y == 200  # 300 - 100
    assert rect.width == 200  # 0.5 scaled to box width 400
    assert rect.height == 100  # 0.25 scaled to box height 400
    assert rect.rounded == 10
    assert "boxed-bottom-center-rect" in rect.classes


def test_build_background_rectangle():
    box = Box(
        top=0,
        right=800,
        bottom=600,
        left=0,
    )
    rect = build_background_rectangle(
        box=box,
        classes=["background-rect"],
    )

    assert rect.top_left.x == 0
    assert rect.top_left.y == 0
    assert rect.width == 800
    assert rect.height == 600
    assert rect.rounded == 0
    assert "background-rect" in rect.classes


def test_build_background_rectangle_with_padding():
    box = Box(
        top=0,
        right=800,
        bottom=600,
        left=0,
        padding_top=0.05,
        padding_right=0.05,
        padding_bottom=0.05,
        padding_left=0.05,
    )
    rect = build_background_rectangle(
        box=box,
        classes=["padded-background-rect"],
    )

    assert rect.top_left.x == 0
    assert rect.top_left.y == 0
    assert rect.width == 800
    assert rect.height == 600
    assert rect.rounded == 0
    assert "padded-background-rect" in rect.classes


@fixture
def rect1() -> Rectangle:
    return Rectangle(
        top_left=Point(x=10, y=20),
        width=200,
        height=100,
        rounded=5,
        classes=["test-rect"],
    )


def test_top(rect1: Rectangle):
    assert rect1.top == 20


def test_right(rect1: Rectangle):
    assert rect1.right == 210


def test_bottom(rect1: Rectangle):
    assert rect1.bottom == 120


def test_left(rect1: Rectangle):
    assert rect1.left == 10


def test_render():
    rect = Rectangle(
        top_left=Point(x=10, y=20),
        width=200,
        height=100,
        classes=["test-rect"],
        styles=["test-style: value;"],
        attributes={"data-test": "rectangle"},
    )

    result = rect.render()
    expected = """<rect 
    class="test-rect" 
    x="10.0" 
    y="20.0" 
    width="200.0" 
    height="100.0" 
    rx="0.0" 
    ry="0.0" 
    style="test-style: value;"
    data-test="rectangle"
/>"""

    assert result == expected



from typing import Callable, Annotated
from pathlib import Path

from newsflash.app import App, Page
from newsflash.widgets import HTML, Button, Select, ValueDisplay, Notifications
from newsflash.widgets.widgets import Widget, BodyParam
from newsflash.svg.element import TemplateParam

from .all_tests import get_all_tests, get_next_failing_test, get_test_by_name


class OldSnapshotHTML(HTML):
    id: Annotated[str, TemplateParam()] = "old-snapshot-html"
    html_content: Annotated[str, TemplateParam()] = get_next_failing_test().load_rendered()


class NewSnapshotHTML(HTML):
    id: Annotated[str, TemplateParam()] = "new-snapshot-html"
    html_content: Annotated[str, BodyParam(), TemplateParam()] = get_next_failing_test().render()


class TestSelect(Select):
    id: Annotated[str, TemplateParam()] = "test-select"
    options: Annotated[list[str], TemplateParam()] = list(get_all_tests().keys())
    default: Callable[[], str] = lambda: get_next_failing_test().name

    def on_select(
        self,
        old_snapshot: OldSnapshotHTML,
        new_snapshot: NewSnapshotHTML,
        test_description: "TestDescription",
        approve_button: "ApproveButton",
    ) -> list[Widget]:
        assert self.selected is not None
        test_class = get_test_by_name(self.selected)

        old_snapshot.html_content = test_class.load_rendered()
        new_snapshot.html_content = test_class.render()
        test_description.value = test_class.description

        if test_class.passes_test():
            approve_button.label = "Snapshots Already Match"
            approve_button.disabled = True
        else:
            approve_button.label = "Approve Snapshot"
            approve_button.disabled = False

        return [old_snapshot, new_snapshot, test_description, approve_button]


class NextChangedSnapshotButton(Button):
    id: Annotated[str, TemplateParam()] = "next-changed-snapshot-button"
    label: Annotated[str, TemplateParam()] = "Next Changed Snapshot"

    def on_click(
        self,
        test_select: TestSelect,
        test_description: "TestDescription",
        old_snapshot: OldSnapshotHTML,
        new_snapshot: NewSnapshotHTML,
        approve_button: "ApproveButton",
    ) -> list[Widget]:
        next_failing_test = get_next_failing_test()

        test_select.selected = next_failing_test.name
        return [
            test_select,
            *test_select.on_select(
                old_snapshot=old_snapshot,
                new_snapshot=new_snapshot,
                test_description=test_description,
                approve_button=approve_button,
            ),
        ]


class TestDescription(ValueDisplay):
    id: str = "test-description"
    label: str = "Test Description"
    value: str = get_next_failing_test().description


class ApproveButton(Button):
    id: Annotated[str, TemplateParam()] = "approve-button"
    label: Annotated[str, TemplateParam()] = (
        "Approve Snapshot"
        if not get_next_failing_test().passes_test()
        else "Snapshots Already Match"
    )
    disabled: Annotated[bool, TemplateParam()] = get_next_failing_test().passes_test()

    def on_click(
        self,
        test_select: TestSelect,
        old_snapshot: OldSnapshotHTML,
        new_snapshot: NewSnapshotHTML,
        notifications: Notifications,
    ) -> list[Widget]:
        assert test_select.selected is not None
        test_class = get_all_tests()[test_select.selected]

        if not test_class.passes_test(new_snapshot.html_content):
            test_class.save_rendered(new_snapshot.html_content)

            # Update the old snapshot content after approval
            old_snapshot.html_content = test_class.load_rendered()
            self.label = "Snapshots Already Match"
            self.disabled = True
            return [self, old_snapshot]

        else:
            notifications.push("Snapshots already match.")
            return [notifications]


class MainPage(Page):
    id: str = ""
    path: str = "/"
    title: str = "Snapshot Test Viewer"
    template: tuple[str, str] = ("snapshot", "snapshot_test.html")

    def compose(self) -> list[Widget]:
        return [
            TestSelect(),
            NextChangedSnapshotButton(),
            OldSnapshotHTML(),
            NewSnapshotHTML(),
            TestDescription(),
            ApproveButton(),
            Notifications(),
        ]


app = App(
    pages=[MainPage()],
    template_folders=[
        ("snapshot", Path(__file__).parent / "templates"),
    ],
)



from pytest import fixture, approx

from newsflash.svg.box import Box, scale_x_to_box, scale_y_to_box


@fixture
def default_box() -> Box:
    return Box(
        top=0,
        right=100,
        bottom=100,
        left=0,
        top_y_value=0.0,
        right_x_value=1.0,
        bottom_y_value=1.0,
        left_x_value=0.0,
    )


def test_simple(default_box):
    new_x = scale_x_to_box(0.4, default_box)
    assert new_x == 40

    new_y = scale_y_to_box(0.6, default_box)
    assert new_y == 60


def test_reversed_x(default_box):
    default_box.right_x_value = 0.0
    default_box.left_x_value = 1.0

    new_x = scale_x_to_box(0.4, default_box)
    assert new_x == 60

    new_y = scale_y_to_box(0.6, default_box)
    assert new_y == 60


def test_reversed_y(default_box):
    default_box.top_y_value = 1.0
    default_box.bottom_y_value = 0.0

    new_x = scale_x_to_box(0.4, default_box)
    assert new_x == 40

    new_y = scale_y_to_box(0.6, default_box)
    assert new_y == 40


def test_x_padding_same_left_right(default_box):
    default_box.padding_left = 0.1
    default_box.padding_right = 0.1

    new_x = scale_x_to_box(0.1, default_box)
    # 10 for padding and 10 for scaled value, 100 for width plus 20 for padding
    # times 100 for scale correction
    assert new_x == approx((10 + 10) / (100 + 20) * 100)


def test_x_padding_different_left_right_a(default_box):
    default_box.padding_left = 0.1
    default_box.padding_right = 0.2

    new_x = scale_x_to_box(0.3, default_box)
    # 10 for padding and 30 for scaled value, 100 for width plus 30 for padding
    # times 100 for scale correction
    assert new_x == approx((30 + 10) / (100 + 30) * 100)


def test_y_padding(default_box):
    default_box.padding_top = 0.1
    default_box.padding_bottom = 0.1

    new_y = scale_y_to_box(0.1, default_box)
    # 10 for padding and 10 for scaled value, 100 for width plus 20 for padding
    # times 100 for scale correction
    assert new_y == approx((10 + 10) / (100 + 20) * 100)


def test_y_padding_different_top_bottom_a(default_box):
    default_box.padding_top = 0.1
    default_box.padding_bottom = 0.05

    new_y = scale_y_to_box(0.25, default_box)
    # 10 for padding and 25 for scaled value, 100 for width plus 15 for padding
    # times 100 for scale correction
    assert new_y == approx((25 + 10) / (100 + 15) * 100)



from math import floor, log10
from decimal import Decimal
from bisect import bisect_left


def order_of_magnitude(value: float) -> int:
    """Returns the order of magnitude of a given value."""
    if value == 0:
        return 0

    return floor(log10(abs(value)))


def n_significant_figures(value: float) -> int:
    """Returns the number of significant figures in a given float value."""
    d = Decimal(str(value)).quantize(Decimal("0.0000000001")).normalize()
    digits = d.as_tuple().digits
    return len(digits)


def order_of_magnitude_for_significant_figures(value: float) -> int:
    n_sig_figs = n_significant_figures(value)
    oom = order_of_magnitude(value)
    return oom - (n_sig_figs - 1)


STEP_CANDIDATES = [0.25, 0.5, 1, 2, 2.5, 5, 10]


def get_y_label_positions(
    values: list[float] | list[int],
    min_y: float | int | None = None,
    max_y: float | int | None = None,
    divide_by: int = 4,
) -> list[float] | list[int]:
    min_y_to_use = min(values) if min_y is None else min_y
    max_y_to_use = max(values) if max_y is None else max_y

    step = 0
    highest_score = 0.0
    num_steps = 0

    for divide_by in [4, 3, 5]:
        ideal_step = (max_y_to_use - min_y_to_use) / divide_by
        normalized_ideal_step = ideal_step / pow(10, order_of_magnitude(ideal_step))

        idx = bisect_left(STEP_CANDIDATES, normalized_ideal_step)
        next_step = STEP_CANDIDATES[idx]

        score = normalized_ideal_step / next_step
        if score > highest_score:
            highest_score = score
            step = next_step * pow(10, order_of_magnitude(ideal_step))
            num_steps = divide_by

    y_label_positions = []
    current_label = min_y_to_use
    for _ in range(num_steps + 1):
        y_label_positions.append(current_label)
        current_label += step

    return y_label_positions


def smart_round(numbers: list[float]) -> tuple[list[float], int]:
    differences = [numbers[i + 1] - numbers[i] for i in range(len(numbers) - 1)]
    min_difference = min(differences)
    oom = order_of_magnitude(min_difference)
    decimal_places = max(0, -oom)

    return [round(num, decimal_places) for num in numbers], decimal_places


def get_step_decimal_places(numbers: list[float]) -> int:
    differences = [numbers[i + 1] - numbers[i] for i in range(len(numbers) - 1)]
    min_difference = min(differences)
    oom = order_of_magnitude_for_significant_figures(min_difference)
    decimal_places = max(0, -oom)

    return decimal_places



from typing import (
    Any,
    Mapping,
    Type,
    Callable,
    get_type_hints,
    TypeVar,
    TYPE_CHECKING,
    get_origin,
    Annotated,
)
from inspect import signature
from typing_extensions import Self

from newsflash.svg.element import Element, TemplateParam


if TYPE_CHECKING:
    from newsflash.app import Page


class QueryParam:
    query_param_name: str | None = None

    def __init__(self, query_param_name: str | None = None) -> None:
        self.query_param_name = query_param_name

    def get_query_param_name(self) -> str | None:
        return self.query_param_name


class BodyParam:
    body_param_name: str | None = None

    def __init__(self, body_param_name: str | None = None) -> None:
        self.body_param_name = body_param_name

    def get_body_param_name(self) -> str | None:
        return self.body_param_name


class WidgetContainer(Element):
    widget_id: Annotated[str, TemplateParam()]
    hx_include: Annotated[list[str], TemplateParam()] = []
    template: tuple[str, str] = ("widgets", "container.html")


W = TypeVar("W", bound="Widget")


class Widget(Element):
    hx_include: Annotated[list[str], TemplateParam()] = []
    hx_swap_oob: Annotated[bool, TemplateParam()] = False

    children: list["Widget"] = []
    parent: "Widget | None" = None

    body_params: Mapping[str, str] | None = None

    _include_parent: bool = False
    _callback_fn_name: str | None = None

    def compose(self) -> list["Widget"]:
        return []

    def append_widgets(self) -> list["Widget"]:
        return []

    def model_copy(
        self,
        *,
        copy: bool = False,
        update: Mapping[str, Any] | None = None,
        query_params: Mapping[str, list[str]] | None = None,
        body_params: Mapping[str, Any] | None = None,
        parent: "Widget | None" = None,
    ) -> Self:
        if parent is not None:
            self.parent = parent

        if copy:
            new_instance = super().model_copy(deep=True, update=update)
        else:
            new_instance = self
            if update is not None:
                for k, v in update.items():
                    setattr(new_instance, k, v)

        query_parameters: dict[str, Any] = {}
        body_parameters: dict[str, Any] = {}

        for k, v in new_instance.__class__.model_fields.items():
            if len(v.metadata) == 0:
                continue

            query_param = next(
                (m for m in v.metadata if isinstance(m, QueryParam)), None
            )
            body_param = next((m for m in v.metadata if isinstance(m, BodyParam)), None)

            if query_param is not None and query_params is not None:
                query_param_name = query_param.get_query_param_name() or k
                value_from_request = query_params.get(query_param_name, [])

                if len(value_from_request) == 0:
                    continue

                if get_origin(v.annotation) == list:
                    query_parameters[k] = value_from_request
                else:
                    query_parameters[k] = value_from_request[0]

            elif body_param is not None and body_params is not None:
                body_param_name = body_param.get_body_param_name() or k
                body_param_name = new_instance.id + "-" + body_param_name

                body_value = body_params.get(body_param_name, None)
                if body_value is None:
                    continue

                body_parameters[k] = body_value

        for k, v in query_parameters.items():
            setattr(new_instance, k, v)

        for k, v in body_parameters.items():
            setattr(new_instance, k, v)

        children = new_instance.compose()
        for child in children:
            child.model_copy(
                copy=False,
                query_params=query_params,
                body_params=body_params,
                parent=new_instance,
            )

        new_instance.children = children
        new_instance._build_hx_include()

        return new_instance

    def get_all_children(
        self,
        type: Type[W],
    ) -> list[W]:
        children_of_type = [
            widget.model_copy() for widget in self.children if isinstance(widget, type)
        ]

        if len(self.children) == 0 and isinstance(self, type):
            return []

        for child in self.children:
            child_children = child.get_all_children(
                type=type,
            )

            children_of_type.extend(child_children)

        return children_of_type

    def get_child_widget(
        self,
        type: Type[W],
        id: str | None = None,
    ) -> W:
        children_of_type = [
            widget for widget in self.children if isinstance(widget, type)
        ]

        if id is None:
            all_children_of_type = self.get_all_children(
                type=type,
            )
            if len(all_children_of_type) == 1:
                return all_children_of_type[0]
            else:
                # TODO: distinguish in error message between no or multiple
                # child widgets found.
                raise ValueError(
                    f"Multiple or no widgets of type {type} found. Specify an id."
                )

        id_split = id.split("/")

        if len(id_split) == 1:
            for widget in children_of_type:
                if widget.id == id:
                    widget_copy = widget.model_copy()
                    return widget_copy

        remaining_id = "/".join(id_split[1:])

        for widget in children_of_type:
            if widget.id == id_split[0]:
                widget_copy = widget.model_copy()

                return widget_copy.get_child_widget(
                    type=type,
                    id=remaining_id,
                )

        raise ValueError(f"Widget not found: {type} with id {id}")

    @property
    def full_path(self) -> str:
        if self.parent is not None:
            full_path = f"{self.parent.full_path}/{self.id}"
        else:
            full_path = self.id

        full_path = full_path.strip("/")

        return full_path

    @property
    def root_widget(self) -> "Page":
        if self.parent is None:
            # assert isinstance(self, Page), "Root widget must be a Page"
            # TODO: how to assert the correct type here?
            return self  # type: ignore
        else:
            return self.parent.root_widget

    def get_additional_context(self) -> dict[str, Any]:
        additional_context = super().get_additional_context()

        children_instances = [
            child.model_copy(update={"parent": self}) for child in self.children
        ]
        rendered_children = {child.id: child.render() for child in children_instances}

        additional_context.update({"widgets": rendered_children})
        additional_context.update({"widget_list": list(rendered_children.values())})
        additional_context.update({"full_path": self.full_path})

        return additional_context

    def _build_hx_include(self) -> None:
        callback_fn = self._get_callback_fn()
        if callback_fn is None:
            return

        sig = signature(callback_fn)
        parameters = sig.parameters

        type_hints = get_type_hints(callback_fn)

        include_list: list[str] = []
        for param in parameters:
            if param == "args" or param == "kwargs":
                continue

            widget_type = type_hints[param]
            if not issubclass(widget_type, Widget):
                continue

            include_list.append(f"#{widget_type.model_fields['id'].default}")

        if self._include_parent and self.parent is not None:
            include_list.append(f"#{self.parent.id}")

        self.hx_include = include_list

    def _get_callback_fn(self) -> Callable | None:
        callback_fn_name = self._callback_fn_name
        if callback_fn_name is None:
            return None

        callback_fn = getattr(self, callback_fn_name, None)

        assert callback_fn is not None, (
            f"Widget has no callback function '{callback_fn_name}'"
        )

        return callback_fn

    def _get_callback_inputs(self) -> dict[str, "Widget"]:
        callback_fn = self._get_callback_fn()
        assert callback_fn is not None, "Widget has no callback function"

        sig = signature(callback_fn)
        parameters = sig.parameters

        type_hints = get_type_hints(callback_fn)

        input_dict = {}
        for param in parameters:
            widget_type = type_hints.get(param, "Unknown")
            if widget_type == "Unknown" or not issubclass(widget_type, Widget):
                continue

            widget = self.root_widget.get_child_widget(
                type=widget_type,
            )

            input_dict[param] = widget

        return input_dict

    def _call_callback(self) -> list["Widget"]:
        callback_fn = self._get_callback_fn()
        assert callback_fn is not None, "Widget has no callback function"

        widgets_to_render = callback_fn(**self._get_callback_inputs())
        return widgets_to_render

    def _render_update(self) -> str:
        self.hx_swap_oob = True
        return self.render()



import polars as pl

from fontTools.ttLib import TTFont

from newsflash.svg.utils.fonts import lora
from newsflash.svg.element import ElementGroup

from .xy_chart import build_xy_chart
from .axes import (
    AxesConfig,
    build_y_axis_config,
    build_spaced_x_axis_config,
)
from .barchart import build_bars


def build_histogram(
    values: list[float] | list[int],
    num_bins: int,
    width: float,
    height: float,
    title: str,
    font: TTFont = lora,
    title_font_size: int = 32,
    label_font_size: int = 16,
) -> ElementGroup:
    values_hist = pl.Series(values).hist(bin_count=num_bins)
    barchart_elements = ElementGroup()

    xs = values_hist.get_column("breakpoint").to_list()
    ys = values_hist.get_column("count").to_list()

    x_width = (max(xs) - min(xs)) / len(xs)
    x_padding = x_width / 2

    axes = AxesConfig(
        x=build_spaced_x_axis_config(values=xs),
        y=build_y_axis_config(values=ys, min_value=0),
    )

    barchart_elements, chart_box = build_xy_chart(
        axes=axes,
        width=width,
        height=height,
        title=title,
        x_padding=x_padding,
        font=font,
        title_font_size=title_font_size,
        label_font_size=label_font_size,
    )

    bars = build_bars(bars=ys, chart_box=chart_box, xs=xs)
    barchart_elements.extend(bars)

    return barchart_elements



from .charts import BarChart, LineChart, Histogram
from .inputs import Input, TextArea, Button, Select
from .text import ValueDisplay, Notifications, Paragraph
from .list import List, Grid
from .html import HTML
from .widgets import QueryParam, BodyParam

__all__ = [
    "BarChart",
    "LineChart",
    "Histogram",
    "Input",
    "TextArea",
    "Button",
    "Select",
    "ValueDisplay",
    "Notifications",
    "Paragraph",
    "List",
    "Grid",
    "HTML",
    "QueryParam",
    "BodyParam",
]



from typing import Annotated

from newsflash.svg.element import ElementGroup, TemplateParam


class SVG(ElementGroup):
    template: tuple[str, str] = ("svg", "svg.svg")
    width: Annotated[float, TemplateParam()]
    height: Annotated[float, TemplateParam()]
    hx_swap_oob: Annotated[bool, TemplateParam()] = False



from typing import Any, TypeVar, Generic, Self, Callable, Mapping

from .widgets import Widget


W = TypeVar("W", bound=Widget)


class List(Widget, Generic[W]):
    """A widget that displays a list of items."""

    item_type: type[W]
    children: list[W] | None = None
    default: Callable[[], list[W]] | None = None
    template: tuple[str, str] | None = ("widgets", "list.html")

    include_in_context: set[str] = {"id", "hx_include", "hx_swap_oob"}

    def model_copy(
        self,
        *,
        copy: bool = False,
        update: Mapping[str, Any] | None = None,
        query_params: Mapping[str, list[str]] | None = None,
        body_params: Mapping[str, Any] | None = None,
        parent: Widget | None = None,
    ) -> Self:
        new_instance = super().model_copy(
            copy=copy,
            update=update,
            query_params=query_params,
            body_params=body_params,
            parent=parent,
        )

        if new_instance.children is None:
            if new_instance.default is not None:
                new_instance.children = new_instance.default()
            else:
                new_instance.children = []

        return new_instance

    def get_list_item_by_id(self, item_id: str) -> W | None:
        if self.children is None:
            return None
        for item in self.children:
            if item.id == item_id:
                return item
        return None

    def get_list_item_by_index(self, index: int) -> W | None:
        if self.children is None:
            return None
        if 0 <= index < len(self.children):
            return self.children[index]
        return None

    @property
    def num_items(self) -> int:
        """Returns the number of items in the list."""
        if self.children is None:
            return 0
        return len(self.children)

    def set_items(self, items: list[W]) -> None:
        self.children = items

    def append_item(self, item: W) -> Self:
        assert self.children is not None, "Items list is not initialized."
        self.children.append(item)
        return self

    def filter_items(self, filter_fn: Callable[[W], bool]) -> Self:
        assert self.children is not None, "Items list is not initialized."
        filtered_items = [item for item in self.children if filter_fn(item)]
        return self.model_copy(update={"items": filtered_items})

    def get_additional_context(self) -> dict[str, Any]:
        assert self.children is not None, "Items list is not initialized."
        for child in self.children:
            child.parent = self

        # TODO: figure out why this assignment complains
        self.children = self.children  # type: ignore

        return super().get_additional_context()

    def get_child_widget(
        self, type: type[W], id: str, body_params: Mapping[str, str] | None = None
    ) -> W:
        child_type = self.item_type
        assert issubclass(child_type, type), (
            f"Child type {child_type} is not a subclass of {type}"
        )

        item_id_split = id.split("/")

        widget_instance = child_type(
            id=item_id_split[0],
            body_params=body_params,
            parent=self,
        )

        if len(item_id_split) == 1:
            return widget_instance
        else:
            return widget_instance.get_child_widget(
                type=type,
                id="/".join(item_id_split[1:]),
            )


class Grid(List[W]):
    """A widget that displays a grid of items."""

    template: tuple[str, str] | None = ("widgets", "grid.html")
    num_columns: int = 2

    include_in_context: set[str] = {"id", "hx_include", "hx_swap_oob", "num_columns"}



from typing import TYPE_CHECKING

from fastapi import Request
from newsflash.templates.templates import template_registry
from newsflash.widgets.widgets import Widget
from newsflash.endpoints.parsers import parse_request_values, _parse_url_path

if TYPE_CHECKING:
    from newsflash.app import Page, App


def get_page_callback(page: "Page", theme_stylesheet: str):
    async def page_endpoint(request: Request) -> str:
        query_params = request.query_params
        query_params_dict: dict[str, list[str]] = {}
        for k in query_params.keys():
            query_params_dict[k] = query_params.getlist(k)

        page_copy = page.model_copy(
            copy=True,
            query_params=query_params_dict,
        )

        rendered_content = page_copy.render()

        page_template = template_registry.get_template("widgets", "index.html")
        return page_template.render(
            request=request,
            title=page_copy.title,
            content=rendered_content,
            theme_stylesheet=theme_stylesheet,
        )

    return page_endpoint


def get_callback_endpoint(widget_id: str, app: "App"):
    async def callback_endpoint(request: Request) -> str:
        body = await request.form()
        headers = request.headers

        url_path = _parse_url_path(headers)
        request_values = parse_request_values(body, headers)

        widget = app.get_widget(
            path=url_path,
            type=Widget,
            id=widget_id,
            body_params=request_values,
            # request_values=request_values,
        )

        widgets_to_render = widget._call_callback()

        rendered_widgets = []
        assert isinstance(widgets_to_render, list), (
            "Callback must return a list of widgets"
        )
        for widget in widgets_to_render:
            assert isinstance(widget, Widget), "Callback must return a list of widgets"
            widget.hx_swap_oob = True
            rendered_widgets.append(widget._render_update())

        return "\n".join(rendered_widgets)

    return callback_endpoint



from pytest import fixture
from unittest.mock import patch

from newsflash.svg.utils import Point
from newsflash.svg.elements import Text, build_text
from newsflash.svg.box import Box


@fixture
def text_pos() -> Point:
    return Point(x=200, y=100)


@fixture
def text_str() -> str:
    return "Test"


@fixture
def text_font_size() -> int:
    return 16


@patch("newsflash.svg.elements.text.get_text_width", return_value=100.0)
def test_build_text(get_text_width_mock, text_pos, text_str, text_font_size):
    text = build_text(
        pos=text_pos,
        text=text_str,
        font=None,  # type: ignore because get_text_width is mocked
        font_size=text_font_size,
        horizontal_align="left",
        classes=["test-class"],
        styles=["fill: black;"],
    )

    assert text.pos == text_pos
    assert text.text == text_str
    assert text.width == 100.0
    assert text.font_size == text_font_size
    assert text.classes == ["test-class"]
    assert text.styles == ["fill: black;", "font-size: 16px;"]


@patch("newsflash.svg.elements.text.get_text_width", return_value=100.0)
def test_build_text_with_box(get_text_width_mock, text_str, text_font_size):
    box = Box(
        top=0,
        right=800,
        bottom=600,
        left=0,
    )

    text = build_text(
        pos=Point(x=0.5, y=0.5),  # Relative position
        text=text_str,
        font=None,  # type: ignore because get_text_width is mocked
        font_size=text_font_size,
        horizontal_align="left",
        box=box,
    )

    assert text.pos == Point(x=400.0, y=300.0)  # Scaled to box
    assert text.text == text_str


@patch("newsflash.svg.elements.text.get_text_width", return_value=100.0)
def test_build_text_align_center(
    get_text_width_mock, text_pos, text_str, text_font_size
):
    text = build_text(
        pos=text_pos,
        text=text_str,
        font=None,  # type: ignore because get_text_width is mocked
        font_size=text_font_size,
        horizontal_align="center",
    )

    assert text.pos == Point(x=150.0, y=100)  # Adjusted for center alignment


@patch("newsflash.svg.elements.text.get_text_width", return_value=100.0)
def test_build_text_align_right(
    get_text_width_mock, text_pos, text_str, text_font_size
):
    text = build_text(
        pos=text_pos,
        text=text_str,
        font=None,  # type: ignore because get_text_width is mocked
        font_size=text_font_size,
        horizontal_align="right",
    )

    assert text.pos == Point(x=100.0, y=100)  # Adjusted for right alignment


@patch("newsflash.svg.elements.text.get_text_width", return_value=100.0)
def test_build_text_align_top(get_text_width_mock, text_pos, text_str, text_font_size):
    text = build_text(
        pos=text_pos,
        text=text_str,
        font=None,  # type: ignore because get_text_width is mocked
        font_size=text_font_size,
        vertical_align="top",
    )

    assert (
        text.pos == Point(x=200.0, y=116.0)  # plus font size
    )  # Adjusted for top alignment


@patch("newsflash.svg.elements.text.get_text_width", return_value=100.0)
def test_build_text_align_middle(
    get_text_width_mock, text_pos, text_str, text_font_size
):
    text = build_text(
        pos=text_pos,
        text=text_str,
        font=None,  # type: ignore because get_text_width is mocked
        font_size=text_font_size,
        vertical_align="middle",
    )

    assert (
        text.pos == Point(x=200.0, y=104.0)  # plus a quarter font size
    )  # Adjusted for top alignment


@patch("newsflash.svg.elements.text.get_text_width", return_value=100.0)
def test_build_text_align_bottom(
    get_text_width_mock, text_pos, text_str, text_font_size
):
    text = build_text(
        pos=text_pos,
        text=text_str,
        font=None,  # type: ignore because get_text_width is mocked
        font_size=text_font_size,
        vertical_align="bottom",
    )

    assert (
        text.pos == Point(x=200.0, y=96.0)  # minus a quarter font size
    )  # Adjusted for top alignment


@fixture
def test_text() -> Text:
    return Text(
        pos=Point(x=100, y=200),
        text="Hello, World!",
        width=150,
        font_size=24,
    )


def test_text_top(test_text: Text):
    assert test_text.top == 176  # 200 - 24


def test_text_right(test_text: Text):
    assert test_text.right == 250  # 100 + 150


def test_text_bottom(test_text: Text):
    assert test_text.bottom == 208  # 200 + 24 / 3


def test_text_left(test_text: Text):
    assert test_text.left == 100


def test_text_height(test_text: Text):
    assert test_text.height == 32  # 24 * 4 / 3


def test_text_render():
    text = Text(
        pos=Point(x=50, y=75),
        text="Sample Text",
        width=120,
        font_size=18,
        classes=["sample-class"],
        styles=["fill: red;"],
        attributes={"data-info": "example", "aria-label": "sample text"},
    )

    result = text.render()
    expected = """<text 
    class="sample-class" 
    x="50.0" 
    y="75.0" 
    style="fill: red;"
    data-info="example"
    aria-label="sample text"
>
    Sample Text
</text>"""
    assert result == expected
